#include "stdafx.h"
//#include "cgalinit.h"
#include "CalculatePockets.h"
#include "Bridge.h"
#include "RemoveDuplicateVertices.h"
#include "TestObjectGenerator.h"

#include <CGAL/IO/Polyhedron_iostream.h>
#include <fstream>
#include <CGAL/IO/Polyhedron_VRML_1_ostream.h> 
#include <CGAL/IO/Verbose_ostream.h>
#include <CGAL/IO/VRML_2_ostream.h>
#include <CGAL/IO/File_writer_inventor.h>

#include "javaview_stream.h"
#include "jv_writer.h"

std::vector<Point_3> points;
Polyhedron_3 polyhedron;
Polyhedron_3 originalPolyhedron; 
std::list<Bridge *> bridgeEdgesList; 
Bridge * longestBridge ; 

//Begin GLUT headers
//#include <glut.h>
//END GLUT headers

//Begin Internal headers
#include "gui.h"
//End Internal Headers

struct Plane_equation {
    template <class Facet>
    typename Facet::Plane_3 operator()( Facet& f) {
        typename Facet::Halfedge_handle h = f.halfedge();
        typedef typename Facet::Plane_3  Plane;
        return Plane( h->vertex()->point(),
                      h->next()->vertex()->point(),
                      h->next()->next()->vertex()->point());
    }
};


/**
 * Test code to use third party lib to output 
 * a CGAL polyhedron to a javaview compatible file.
 * ----------------------------------------------------------
 * Intention being to one day be able to display polyhedrons
 * on website easily using javaview applet (instead of 
 * resorting to vrml and required third party viewer). 
 */
void writePolygonToJVX(char* pFilename, Polyhedron_3 *pConvexHull) { 
	//use the Javaview_writer for output on  
	//FIRST A FILE/////////////////////////////////////////////
	std::ofstream oFile(pFilename, std::ios::out);
	CGAL::Javaview_writer<std::ofstream> jvw(oFile);
	jvw.set_title("Convex Hull");

	//The description of multiple geometries must be given between a
	//write_header() and a write_footer().
	jvw.write_header();
	//BEGINNING OF GEOMETRIES-----------------------------------------------------//
	CGAL::Color meshColor(100, 100, 100);
	//first a polysurf
	jvw.set_color(meshColor);
	jvw.set_geometry_name("cgal poly paraboloid");
	jvw.begin_geometry();

	polyhedron_javaview_writer(jvw, *pConvexHull);
	jvw.end_geometry();
	//END OF GEOMETRIES-----------------------------------------------------//
	jvw.write_footer();
}


void writeBridgePocketModelToJVX(char *pFilename, Polyhedron_3 *pPoly) { 
	//use the Javaview_writer for output on  
	//FIRST A FILE/////////////////////////////////////////////
	std::ofstream oFile(pFilename, std::ios::out);
	CGAL::Javaview_writer<std::ofstream> jvw(oFile);
	jvw.set_title("Bridges and Pockets");

	//The description of multiple geometries must be given between a
	//write_header() and a write_footer().
	jvw.write_header();
	

	//first draw mesh
	CGAL::Color meshColor(100, 100, 100);
	//first a polysurf
	jvw.set_color(meshColor);
	jvw.set_geometry_name("Polyhedral Mesh");
	jvw.begin_geometry();

	polyhedron_javaview_writer(jvw, *pPoly);
	jvw.end_geometry();
	//end of drawing mesh


	//BEGIN DRAWING OF BRIDGES AND POCKETS
	CGAL::Color bridgeColor(0, 0, 255); 
	CGAL::Color pocketColor(0, 255, 0); 

	jvw.set_color(bridgeColor);
	jvw.set_geometry_name("bridges");
	jvw.begin_geometry();

	//first write out the points of all bridges
	std::list<Bridge *>::iterator cur(bridgeEdgesList.begin());
	*jvw.get_stream() << "         <pointSet dim=\"3\" point=";
	*jvw.get_stream() << "\"hide\">\n";
	*jvw.get_stream()  << "            <points>\n";
	while (cur != bridgeEdgesList.end()) {
		Bridge * curBridge = (*cur); 

		Point_3 p1 = curBridge->oPolyVertexH1->point(); 
		Point_3 p2 = curBridge->oPolyVertexH2->point(); 
		*jvw.get_stream() << "               <p>"
				 << ::CGAL::to_double( p1[0]) << " "
				 << ::CGAL::to_double( p1[1]) << " "
				 << ::CGAL::to_double( p1[2]) 
				 << "</p>\n";
		*jvw.get_stream() << "               <p>"
				 << ::CGAL::to_double( p2[0]) << " "
				 << ::CGAL::to_double( p2[1]) << " "
				 << ::CGAL::to_double( p2[2]) 
				 << "</p>\n";

		cur++; 
	}
	jvw.write_color();

	*jvw.get_stream() << "            </points>\n"
				<< "         </pointSet>\n";
	//END writing of bridges' points

	//BEGIN writing of bridges lines (using the pointset just defined)
	cur = bridgeEdgesList.begin();
	*jvw.get_stream() << "         <lineSet line=\"show\">\n"
			       << "            <lines>\n";
	int curLineIndex=0; 
	while (cur != bridgeEdgesList.end()) {
		Bridge * curBridge = (*cur); 

		Point_3 p1 = curBridge->oPolyVertexH1->point(); 
		Point_3 p2 = curBridge->oPolyVertexH2->point(); 
		*jvw.get_stream() << "               <l>" << curLineIndex << " " << (curLineIndex+1) << "</l>\n";
		curLineIndex+=2; 
		cur++; 
	}
	*jvw.get_stream() << "<thickness>3</thickness>"; 
	jvw.write_color();
	*jvw.get_stream() << "            </lines>\n"
				<< "         </lineSet>\n";
	//END WRITING edges of bridges

	//finish geometry describing bridges
	jvw.end_geometry();
	//DONE WRITING bridges
	



	//BEGIN WRITING Pockets
	jvw.set_color(pocketColor);
	char pocketName[1024]; 
	int curPocket = 0; 
	cur = bridgeEdgesList.begin(); 
	while (cur != bridgeEdgesList.end()) {
		Bridge * curBridge = (*cur); 

		std::list<VERTEX_TYPE>::iterator pocketIt(curBridge->pocketVertices.begin());
		Vertex_handle prevVertex, curVertex;
		sprintf(pocketName, "pocket%d\0", curPocket++); 
		jvw.set_geometry_name(pocketName);
		jvw.begin_geometry();
	
		int numPocketPoints = 2; //start at two 
		/**
		 *Version below includes end points (the two nodes of the bridge)
		 */
		//STEP 1: create point set of all points of the pocket
		*jvw.get_stream() << "         <pointSet dim=\"3\" point=\"hide\">\n";
		*jvw.get_stream()  << "            <points>\n";
		Point_3 p1 = curBridge->oPolyVertexH2->point(); 
		*jvw.get_stream() << "               <p>"
				<< ::CGAL::to_double( p1[0]) << " "
				<< ::CGAL::to_double( p1[1]) << " "
				<< ::CGAL::to_double( p1[2]) 
				<< "</p>\n";
		while (pocketIt != curBridge->pocketVertices.end()) {
			curVertex = *pocketIt; 
			//Point_3 p1 = prevVertex->point(); 
			Point_3 p2 = curVertex->point(); 
			/*
			*jvw.get_stream() << "               <p>"
				<< ::CGAL::to_double( p1[0]) << " "
				<< ::CGAL::to_double( p1[1]) << " "
				<< ::CGAL::to_double( p1[2]) 
				<< "</p>\n";
			*/
			*jvw.get_stream() << "               <p>"
				<< ::CGAL::to_double( p2[0]) << " "
				<< ::CGAL::to_double( p2[1]) << " "
				<< ::CGAL::to_double( p2[2]) 
				<< "</p>\n";
			 
			//prevVertex = curVertex; 
			pocketIt++; 
			numPocketPoints++; 
		}
		p1 =  curBridge->oPolyVertexH1->point(); 
		*jvw.get_stream() << "               <p>"
				<< ::CGAL::to_double( p1[0]) << " "
				<< ::CGAL::to_double( p1[1]) << " "
				<< ::CGAL::to_double( p1[2]) 
				<< "</p>\n";
		*jvw.get_stream() << "            </points>\n" << "         </pointSet>\n";
		//END step 1

		//STEP 2: write edges of the pocket (just connecting all points in order)
		*jvw.get_stream() << "         <lineSet line=\"show\">\n" 
			<< "            <lines>\n";
		*jvw.get_stream() << "               <l>";
		//one "line" for each pocket, which is just listing of all its points defined above
		for(int i = 0; i!=numPocketPoints; i++ ) { 
			*jvw.get_stream()  << i << " ";
		}
		*jvw.get_stream() << "</l>\n";
		jvw.write_color();
		*jvw.get_stream() << "<thickness>3</thickness>"; 
		*jvw.get_stream()	<< "            </lines>\n"
			<< "         </lineSet>\n";

		//finished drawing this pocket for current bridge
		jvw.end_geometry();
		
		//increment to draw next bridge's pocket
		cur++; 
	}
	//END WRITING POCKETS

	//DONE writing all geometry
	jvw.write_footer();
}

/**
 * Debug code to calculate the bridge with the longest
 * chain of pockets. 
 */
void computeLongestPocket(std::list<Bridge *> * pBridgeEdgesList) {
	std::list<Bridge *>::iterator cur(pBridgeEdgesList->begin());
	double maxDistance = 0; //keeps track of the longest bridge seen so far
	//Bridge * longestBridge; 
	longestBridge = NULL; 
	while (cur != pBridgeEdgesList->end()) {
		Bridge * curBridge = (*cur); 
		if (curBridge->pocketLength > maxDistance) { 
			longestBridge = curBridge;
			maxDistance = curBridge->pocketLength; 
		}
		cur++; 
	}
}


/**
 * Loads an input mesh from a file.
 * Note becuase of CGAL, there are MANY restrictions on mesh...
 * best to run through some program that can "fix up" meshes
 * before running this program on it. 
 */
int loadMesh(char *fileName) { 
	std::istream*     p_in = &std::cin;
	std::ifstream inputFile;
	
	
	//inputFile.open("data/big_and_little_rects_unix.off");
	inputFile.open(fileName);
	std::cout << "Loading mesh from " << fileName << " ..."; 

	p_in = &inputFile;

	if ( !*p_in) { 
		std::cerr << "error: cannot open file '"<< fileName
			<< "' for reading." << std::endl;
		exit(1);
	}

	//read mesh from file and load into originalPolyhedron
	CGAL::set_ascii_mode(inputFile); 

	(*p_in) >> originalPolyhedron;

	if ( !*p_in) { 
		std::cerr << "offview: read error: while reading file '"
			<< "filename" << "'." << std::endl;
		exit(1);
	}

	std::cout << " success!" << std::endl; 
	return 1; 
}

void calculateBridges(Polyhedron_3 * convexHullPolyhedron) { 
	Edge_iterator it;
	Vertex_iterator polyVertexIt; 
	for(it = convexHullPolyhedron->edges_begin(); it != convexHullPolyhedron->edges_end(); it++)
	{
		Halfedge_handle he = it;
		//convex hull points
		const Point_3& chp1 =	he->opposite()->vertex()->point();
		const Point_3& chp2 =	he->vertex()->point();
		#ifdef DEBUG_FIND_BRIDGES
		std::cout << "CH Edge " << chp1 << " --> " << chp2 << std::endl; 
		#endif
		if (true) {
			//find first vertex in the original object
			for (polyVertexIt = originalPolyhedron.vertices_begin(); 
				polyVertexIt !=originalPolyhedron.vertices_end() && polyVertexIt->point()!=(chp1); // && polyVertexIt->point().operator!=(chp2); 
				polyVertexIt++); 
			//note below should never be true (means point on convex hull which wasn't in originalPolyhedron)
			if (polyVertexIt==originalPolyhedron.vertices_end()) { 
				//#ifdef DEBUG_FIND_BRIDGES
				std::cout << "\tError finding corresponding original point" << std::endl; 
				//#endif
				exit(0); 
			}

			//check edges of original polyhedron connected to the current vertex (polyVertexIt) to check if they
			else { 
				Halfedge_around_vertex_circulator origPolyEdgeIt = polyVertexIt->vertex_begin();
				bool breakLoop = false; 
				Bridge * bridge = NULL; 
				do { 
					//first locate the currrent point 
					//Point_3 *curPoint, *otherPoint; //current point is the one found in both convex hull and original poly, other may or may not be on both. 
					const Point_3& origp1 = origPolyEdgeIt->opposite()->vertex()->point();
					const Point_3& origp2 = origPolyEdgeIt->vertex()->point();
					#ifdef DEBUG_FIND_BRIDGES
					std::cout << "\tEdge " << origPolyEdgeIt->opposite()->vertex()->point() << " --> " << origPolyEdgeIt->vertex()->point() << std::endl; 
					#endif
					if ( (origp1==(chp1) && origp2==(chp2)) || (origp1==(chp2) && origp2==(chp1))) { 
						//std::cout << "\tmatch found: " << origp1 << "->" << origp2 << " == " << chp1 << "->" << chp2 << std::endl;
						breakLoop = true; 
						break; 
					}
				} while (!breakLoop && ++origPolyEdgeIt!=polyVertexIt->vertex_begin());
				//if we iterated through all of the edges of the point without finding one that matches the edge from the convex hull, then it is a bridge (minus triangulation issues)
				if (!breakLoop) { 
					#ifdef DEBUG_FIND_BRIDGES
					std::cout << "\tbridge detected" << std::endl;
					#endif
					//FIXME: note that right now we do not have a handle to the other vertex on THE ORIGINAL polyhedron, so the second vertex currently stored is the other vertex on the CONVEX hull (both have same point, but not same vertex). 
					bridge = new Bridge(polyVertexIt, he->vertex(), he); 	
					bridgeEdgesList.push_back(bridge); //add the edge 				
				}
			}
		}
	}
}


int main(int argc, char **argv)
{
	std::cout << "Running Decomposition of Convex Polyhedron" << std::endl;
	
	if (argc<2) { 
		printf("Error: invalid # of parameters\n");
		printf("Usage: %s (input OFF file)\n", argv[0]); 
	}

	loadMesh(argv[1]); 

	Polyhedron_3 * p = &originalPolyhedron;

	polyhedron = originalPolyhedron; 

	// define object to hold convex hull
	CGAL::Object ch_object;
	Polyhedron_3 convexHullPolyhedron; 

	// compute convex hull
	std::cout << "Computing convex hull of input point cloud... "; 
	CGAL::convex_hull_3(p->points_begin(), p->points_end(), convexHullPolyhedron);
	std::cout << " finished!" << std::endl; 
	//polyhedron = convexHullPolyhedron;
	
	//writePolygonToJVX("data/inputmesh.jxv", p); 
	//writePolygonToJVX("data/convexhull.jxv", &convexHullPolyhedron); 

	//TODO: remove/combine vertices which have same point. Need method of doing so without breaking CGAL restrictions on mesh! 

	//OPTIMIZATION: should find way to modify convex_hull_3 algorithm to link points or vertices in original polyhedron and the convex hull to simplify bridge/pocket location (instead of having to search over all points each time)
	
	//find all edges in convex hull that are not in polyhedron
	std::cout << "Calculating Bridges (edges on CH not on P)..." << std::endl; 
	calculateBridges(&convexHullPolyhedron); 
	std::cout << "Finished calculating bridges" << std::endl;
	
	if (true) { 
		//iterate over/printout bridges and remove bridges bordering coplanar facets
		/*std::list<Bridge *>::iterator cur(bridgeEdgesList.begin());
		while (cur != bridgeEdgesList.end()) {
			Bridge *curBridge = (*cur); 

			Halfedge_handle he = (curBridge->convexHullHalfEdge);
			Halfedge_handle ohe = he->opposite(); //half-edge of adjacent facet

			//see if the two facets bounded by the edge are coplanar, and if so remove them ( shouldn't they infact be grouped together?)
			Plane_3 plane1(he->vertex()->point(),
				he->next()->vertex()->point(),
				he->next()->next()->vertex()->point());
			Plane_3 plane2(ohe->vertex()->point(),
				ohe->next()->vertex()->point(),
				ohe->next()->next()->vertex()->point());
			if (plane1.operator==(plane2)) { 
				//std::cout << "\tedge divides coplanar facets" << std::endl; 
				cur++; 
				bridgeEdgesList.remove(curBridge); 
			} else { 
				cur++; 
			}
		}*/

		/*
		* Locate pockets; iterate over bridges, and for each bridge, calculate the 
		* shortest set of edges of the original polyhedron which connect the two
		* vertices of the bridge. 
		*/
		//cur = bridgeEdgesList.begin();
		CCalculatePockets::calcPockets(&originalPolyhedron, &convexHullPolyhedron, &bridgeEdgesList); 
	}

	//draw bridges adn pcokets to model
	//writeBridgePocketModelToJVX("data/pocketsandbridges.jvx", p); 

	//compute longest bridge (bridge with longest set of pockets)
	if (true) { 
		computeLongestPocket(&bridgeEdgesList); 
	}

	if (true) { 
		
	}

	initOpenGL(argc, argv); 
	return 0;
}


//below are code snippets kept around strictly for reference in case they are needed later. 


	/*
	 * Below is code snippet to iterate over all facets of a polygon, and within each facet
	 * iterate over all of its half edges. 
	 */
	/*
	for ( Facet_iterator pFacet = p->facets_begin(); pFacet != p->facets_end(); ++pFacet) {
		Halfedge_facet_circulator j = pFacet->facet_begin();
		do {
			std::cout << j->vertex()->point() << std::endl; 
        } while ( ++j != pFacet->facet_begin());
	}
	*/
	//End of facet/halfedge iterator snippet

	/*
	Halfedge_iterator it;
		for(it = p->halfedges_begin(); it != p->halfedges_end(); it++)
		{
			Halfedge_handle he = it;
			const Point_3& p1 =	he->opposite()->vertex()->point();
			const Point_3& p2 =	he->vertex()->point();
			//::glVertex3d(p1[0],p1[1],p1[2]);
			//::glVertex3d(p2[0],p2[1],p2[2]);
		}
		*/ 

	/*
	for ( Facet_iterator pFacet = p->facets_begin(); pFacet != p->facets_end(); ++pFacet) {
        std::cout << pFacet-> << std::endl;
		Point_3 point = v->point(); 
		double x,y,z,w;

		x =  CGAL::to_double(point.x());
		y =  CGAL::to_double(point.y());
		z =  CGAL::to_double(point.z()); 
		 std::cout <<x << "," << y << "," << z << std::endl;
		
		glVertex3d(x,y,z);	
	}
	 */ 

//write off as VRML file
	/*
	std::ofstream     out;
	out.open("output.wrl"); 
	CGAL::VRML_1_ostream os(out);
    os << originalPolyhedron;
*/ 

/*
		* Calculate Bounding box of points
		* Reference: http://www.cgal.org/Manual/3.3/doc_html/cgal_manual/Principal_component_analysis/Chapter_main.html
		*
		*/
		/*
		K::Iso_cuboid_3 c3 = CGAL::bounding_box(p->points_begin(), p->points_end());
		std::cout << c3 << std::endl; 
		*/ 


/*
//Print out vertices of imported polyhedron
	if (false) { 
	for ( Vertex_iterator v = p->vertices_begin(); v != p->vertices_end(); ++v) {
       // std::cout << v->point() << std::endl;
		Point_3 point = v->point(); 
		double x,y,z,w;

		x =  CGAL::to_double(point.x());
		y =  CGAL::to_double(point.y());
		z =  CGAL::to_double(point.z()); 
		 std::cout <<x << "," << y << "," << z << std::endl;
		
		//glVertex3d(x,y,z);	
	}
	std::cout << std::endl << std::endl; 
	}
	*/ 