#include "stdafx.h"
#include "CalculatePockets.h"

//below was derived from http://www.cgal.org/Manual/3.3/doc_html/cgal_manual/BGL/Chapter_main.html

//define BOOST related
//#include <CGAL/Polyhedron_items_with_id_3.h>
#include <CGAL/boost/graph/graph_traits_Polyhedron_3.h>
#include <CGAL/boost/graph/properties_Polyhedron_3.h>
#include <boost/graph/dijkstra_shortest_paths.hpp>

#include <algorithm>
#include <list>

typedef boost::graph_traits<Polyhedron_3>::vertex_descriptor vertex_descriptor;
typedef boost::graph_traits<Polyhedron_3>::vertex_iterator   vertex_iterator;
typedef boost::graph_traits<Polyhedron_3>::edge_descriptor   edge_descriptor;
//end boost related

CCalculatePockets::CCalculatePockets(void)
{
}

CCalculatePockets::~CCalculatePockets(void)
{
}

/**
 * Apply dijsktras to finding the shortest path on the mesh
 * connecting two nodes (in this case, specifically the two
 * vertices of the bridge). The shortest path (upon the 
 * original mesh) is the "pocket" associated with a particular
 * "bridge" (an edge on the convex hull that isn't also on the
 * original polyhedral mesh). 
 */ 
void dijkstra_test(Polyhedron_3 *srcP, Bridge *pBridge) { 
	Vertex_handle source = pBridge->oPolyVertexH1;
	Vertex_handle target = pBridge->oPolyVertexH2; 

	// associate indices to the vertices using the "id()" field of the vertex.
	vertex_iterator vb, ve;
	int index = 0;

	// boost::tie assigns the first and second element of the std::pair
	// returned by boost::vertices to the variables vb and ve
	//FIXME: below should only need to be done once on the polyhedron, not per call
	for(boost::tie(vb,ve)=boost::vertices(*srcP); vb!=ve; ++vb ){
		vertex_descriptor  vd = *vb;
		vd->id() = index++;
	}

	//vertex_descriptor source =  *boost::vertices(*srcP).first;
	//std::cout << "\nStart dijkstra_shortest_paths at (" << source->point() << ") --> (" << target->point() << ") " << std::endl;

	typedef boost::property_map<Polyhedron_3, boost::vertex_index_t>::type VertexIdPropertyMap;
	VertexIdPropertyMap vertex_index_pmap = get(boost::vertex_index, *srcP);

	// Dijkstra's shortest path needs property maps for the predecessor and distance
	std::vector<vertex_descriptor> predecessor(boost::num_vertices(*srcP));
	boost::iterator_property_map<std::vector<vertex_descriptor>::iterator, VertexIdPropertyMap>
		predecessor_pmap(predecessor.begin(), vertex_index_pmap);

	boost::dijkstra_shortest_paths(*srcP, source, predecessor_map(predecessor_pmap));
	
	//OPTIMIZE: find method to immediately locate target vertex rather than iterate until the points match
	vertex_descriptor targetVD = NULL; 
	vertex_descriptor vd, lastVD;
#define DEBUG_DIJSKTRA
#ifdef DEBUG_DIJSKTRA
	int numMatchingPoints = 0; 
#endif
#ifndef DEBUG_DIJSKTRA
	//origianl
	for(boost::tie(vb,ve)=boost::vertices(*srcP); targetVD==NULL && vb!=ve; ++vb ){
#else
	//run through all nodes
	for(boost::tie(vb,ve)=boost::vertices(*srcP); vb!=ve; ++vb ){
#endif
		vd = *vb;
		if (vd->point()==target->point()) { 
			targetVD = vd; 
#ifdef DEBUG_DIJSKTRA
			numMatchingPoints++; 
#endif
		} 
	}
	if (numMatchingPoints>1) { 
		std::cout << "Error found multiple matches for node!!! (" << numMatchingPoints << std::endl; 
	}
	if (targetVD!=NULL) { 
		//std::cout << "target found: " << (targetVD->id()+1) << " " << targetVD->point() << std::endl;
		//list the vertices of the shortest path connecting source and target on the original polyhedron. 
		vd = get(predecessor_pmap,targetVD);
		lastVD = NULL; 
		/**
		 * Make sure not to get into infinite loop (where predecessor of vd = vd). 
		 */ 
		while (vd!=source && vd !=lastVD) { 
			pBridge->addPocketVertex(vd); 
			//std::cout << "[" << (vd->id()+1) << " - (" << vd->point() << ")]\t"; 
			lastVD = vd; 
			vd = get(predecessor_pmap,vd);
		}; 
	}

	else {
		std::cout << " failed to find point!" << std::endl; 
	}

	//std::cout << "finished Dijkstra's on point" << std::endl; 
}

void CCalculatePockets::calcPockets(Polyhedron_3 *srcP, Polyhedron_3 * cHP,  std::list<Bridge *> * pBridgeEdgesList) {
	std::list<Bridge *>::iterator cur(pBridgeEdgesList->begin());
	//for initial dev, just operate on a single bridge 

	//first assign unique index to each vertex
	// associate indices to the vertices using the "id()" field of the vertex.
	vertex_iterator vb, ve;
	int index = 0;

	// boost::tie assigns the first and second element of the std::pair
	// returned by boost::vertices to the variables vb and ve
	for(boost::tie(vb,ve)=boost::vertices(*srcP); vb!=ve; ++vb ){
		vertex_descriptor  vd = *vb;
		vd->id() = index++;
	}

	while (cur != pBridgeEdgesList->end()) {
		Bridge * curBridge = (*cur); 
		//Vertex_handle origPSrcVertex = curBridge->oPolyVertexH1;
		//dijkstra_test(srcP, curBridge->oPolyVertexH1, curBridge->oPolyVertexH2); 
		dijkstra_test(srcP, curBridge); 
		
		Vertex_handle previousVh = NULL; //handle to previous pocket vertex
		std::list<VERTEX_TYPE>::iterator pocketIt(curBridge->pocketVertices.begin());
		while (pocketIt != curBridge->pocketVertices.end()) {
			Vertex_handle vh = *pocketIt; 
			//calculate concavity (distance between vertex and the bridge)
			//create a segment representing the bridge
			Segment_3 bridgeSegment(curBridge->oPolyVertexH1->point(), curBridge->oPolyVertexH2->point()); 
			double concavity = sqrt(CGAL::squared_distance(vh->point(), bridgeSegment)); 
			if (previousVh!=NULL) { 
				curBridge->pocketLength += CGAL::squared_distance(vh->point(), previousVh->point()); 
			}
			previousVh = vh; 
			//std::cout << vh->id() << ":	" << concavity << std::endl; 
			pocketIt++; 
		}
		//std::cout << curBridge->pocketLength << std::endl; 
		cur++; 
	}

}

/**
 * Calculates the concavity between the bridge and a vertex (point)
 * with the assumption that the vertex must be 
 */
/*
void CalculatePockets::CalculateConcavity(Bridge * pBridge, Vertex_handle vertex) { 

}
*/ 