#include "stdafx.h"
//#include "cgalinit.h"
#include <CGAL/Polyhedron_incremental_builder_3.h>
//#include <CGAL/spatial_sort.h>
#include <CGAL/Orthogonal_k_neighbor_search.h>
#include <CGAL/Search_traits_3.h>

#include "RemoveDuplicateVertices.h"

typedef CGAL::Search_traits_3<K> TreeTraits;
typedef CGAL::Orthogonal_k_neighbor_search<TreeTraits> Neighbor_search;
typedef Neighbor_search::Tree Tree;

//#define Point Point_3

// A modifier creating a triangle with the incremental builder.
template <class HDS>
class Build_Duplicate_Unit_Test : public CGAL::Modifier_base<HDS> {
public:
    Build_Duplicate_Unit_Test() {}
    void operator()( HDS& hds) {
        // Postcondition: `hds' is a valid polyhedral surface.
        CGAL::Polyhedron_incremental_builder_3<HDS> B( hds, true);
        B.begin_surface(5, 4, 6); 
        typedef typename HDS::Vertex   Vertex;
        typedef typename Vertex::Point Point;
        B.add_vertex( Point_3( -1, 0, 0));
        B.add_vertex( Point_3( 1, 0, 0));
        B.add_vertex( Point_3( 0, 0.5, 1));
		B.add_vertex( Point_3( 0, 1, 0));
		B.add_vertex( Point_3( 1, 0, 0));

        B.begin_facet();
        B.add_vertex_to_facet( 0);
        B.add_vertex_to_facet( 1);
        B.add_vertex_to_facet( 3);
        B.end_facet();
		
		B.begin_facet();
        B.add_vertex_to_facet( 1);
        B.add_vertex_to_facet( 0);
        B.add_vertex_to_facet( 2);
        B.end_facet();
		
		B.begin_facet();
        B.add_vertex_to_facet( 3);
        B.add_vertex_to_facet( 4);
        B.add_vertex_to_facet( 2);
        B.end_facet();
		
		//tried 0,3,2: error
		//tried 0,2,3: error
		B.begin_facet();
        B.add_vertex_to_facet( 2);
        B.add_vertex_to_facet( 0);
        B.add_vertex_to_facet( 3);
        B.end_facet();
		
        B.end_surface();
    }
};


/**
 * Generates a test Polyhedron_3 with duplicate vertices
 */
Polyhedron_3 * CRemoveDuplicateVertices::getDupliateTestObject() { 
	Polyhedron_3 * P = new Polyhedron_3(); 
    Build_Duplicate_Unit_Test<HalfedgeDS> triangle;
    P->delegate( triangle);
	return P; 
}



void CRemoveDuplicateVertices::testRemoveDuplicateVertices() { 
	Polyhedron_3 * p = getDupliateTestObject(); 
	//Tree tree;

	for ( Vertex_iterator v = p->vertices_begin(); v != p->vertices_end(); ++v) {
		// std::cout << v->point() << std::endl;
		Point_3 point = v->point(); 
		double x,y,z,w;

		x =  CGAL::to_double(point.x());
		y =  CGAL::to_double(point.y());
		z =  CGAL::to_double(point.z()); 
		std::cout <<x << "," << y << "," << z << std::endl;
	}

	//sort points
	//doesn't compile!
	//spatial_sort(p->points_begin(), p->points_end(), p->traits());

	//end of spatial_sort attempt

	//begin tree-based searching
	/*
	Tree tree(p->points_begin(), p->points_end());
	
	Point_3 query(0,0,0);

	// Initialize the search structure, and search all N points
	const int N = 1;
	Neighbor_search search(tree, query, N);

	// report the N nearest neighbors and their distance
	// This should sort all N points by increasing distance from origin
	std::cout << "Points found" << std::endl; 
	for(Neighbor_search::iterator it = search.begin(); it != search.end(); ++it) {
		std::cout << it->first << " "<< std::sqrt(it->second) << std::endl;
	}
	std::cout <<  std::endl; 
	//
	*/ 

	Tree tree(p->points_begin(), p->points_end());
	for ( Vertex_iterator v = p->vertices_begin(); v != p->vertices_end(); ++v) {
		Point_3 curPoint = v->point(); 
		Neighbor_search search(tree, curPoint, 5);

		//calculate the number of duplicate points
		//note as implemented below, only counts MIN(# of duplicates, 5), need to loop, and either remove points
		int numDuplicates=-1; 
		for(Neighbor_search::iterator it = search.begin(); it != search.end(); ++it) {
			if (it->second == 0) { 
				numDuplicates++; 
			}
			else {
				break;
			}
		}
		std::cout <<  "Point " << curPoint << " has " << numDuplicates << " dupliates" << std::endl; 


		/**
		 * Pseudo-code, for each point
		 *	1) search for the point
		 *  2) iterate over 5 nearest matches 
		 *		all matches represent duplicates, but this doesn't provide duplicate vertices

		 * need point --> vertex
		 */
	}
}

CRemoveDuplicateVertices::CRemoveDuplicateVertices(void)
{
	
}

CRemoveDuplicateVertices::~CRemoveDuplicateVertices(void)
{

}
