#include "stdafx.h"
//#include "cgalinit.h"

#include <stdio.h>
//disables ZPR's debugging
#define NDEBUG
#include "zpr.h"

#include "Bridge.h" 

//Begin GLUT headers
//#include <glut.h>
//END GLUT headers

//DEFINES to control display
#define DRAW_BRIDGES
#define DRAW_POCKETS
#define DRAW_LONGEST_BRIDGE
//end of display #defines


extern std::vector<Point_3> points;
extern Polyhedron_3 polyhedron;
extern std::list<Bridge *> bridgeEdgesList; 
extern Bridge * longestBridge; 

/* Macro for checking OpenGL error state */
#define GLERROR                                                    \
    {                                                              \
        GLenum code = glGetError();                                \
        while (code!=GL_NO_ERROR)                                  \
        {                                                          \
            printf("%s\n",(char *) gluErrorString(code));          \
                code = glGetError();                               \
        }                                                          \
    }
 

//cite: reducetounit and calcNormal functions taken from OpenGL Bible on Books24x7
// Reduces a normal vector specified as a set of three coordinates	
// to a unit normal vector of length 1.	
void ReduceToUnit(float vector[3])	
{	
	float length;	

	// Calculate the length of the vector	
	length = (float)sqrt((vector[0]*vector[0]) +	
		(vector[1]*vector[1]) +	
		(vector[2]*vector[2]));	

	// Keep the program from blowing up by providing an acceptable	
	// value for vectors whose length may be calculated too close to zero.	
	if(length == 0.0f)	
		length = 1.0f;	

	// Dividing each element by the length will result in a	
	// unit normal vector.	
	vector[0] /= length;	
	vector[1] /= length;	
	vector[2] /= length;	
}
void calcNormal(float v[3][3], float out[3])	
{	
	float v1[3],v2[3];	
	static const int x = 0;	
	static const int y = 1;	
	static const int z = 2;	

	// Calculate two vectors from the three points	
	v1[x] = v[0][x] - v[1][x];	
	v1[y] = v[0][y] - v[1][y];	
	v1[z] = v[0][z] - v[1][z];	

	v2[x] = v[1][x] - v[2][x];	
	v2[y] = v[1][y] - v[2][y];	
	v2[z] = v[1][z] - v[2][z];	

	// Take the cross product of the two vectors to get	
	// the normal vector, which will be stored in out[]	
	out[x] = v1[y]*v2[z] - v1[z]*v2[y];	
	out[y] = v1[z]*v2[x] - v1[x]*v2[z];	
	out[z] = v1[x]*v2[y] - v1[y]*v2[x];	

	// Normalize the vector (shorten length to one)	
	ReduceToUnit(out);	
}

void normalize(Vector_3& v)
{
	double d = CGAL::to_double(v.squared_length()); //v.x()*v.x() + v.y()*v.y() + v.z()*v.z();
	d = sqrt(d);
	v = Vector_3(v.x()/d, v.y()/d, v.z()/d);
}

/* Draw axes */
#define RANDOM_FLOAT ((float) rand() / (float) RAND_MAX)
void drawPolyhedron(Polyhedron_3 * p) { 
	
	/*
	glBegin(GL_LINES);
	Halfedge_iterator it;
		for(it = p->halfedges_begin(); it != p->halfedges_end(); it++)
		{
			Halfedge_handle he = it;
			const Point_3& p1 =	he->opposite()->vertex()->point();
			const Point_3& p2 =	he->vertex()->point();
			glVertex3d(CGAL::to_double(p1[0]),CGAL::to_double(p1[1]),CGAL::to_double(p1[2]));
			glVertex3d(CGAL::to_double(p2[0]),CGAL::to_double(p2[1]),CGAL::to_double(p2[2]));
		}
		*/
	float polyhedronColor = .2;
	glColor3f(polyhedronColor, polyhedronColor,polyhedronColor); 
	Vector_3 normalsVector;

	for (Facet_iterator pFacet = p->facets_begin(); pFacet != p->facets_end(); ++pFacet) {
		Halfedge_facet_circulator j = pFacet->facet_begin();
		normalsVector = CGAL::cross_product( 
					  j->next()->vertex()->point() - j->vertex()->point(),
					  j->next()->next()->vertex()->point() - j->next()->vertex()->point());
		
		normalize(normalsVector);
		
		//glColor3f(RANDOM_FLOAT, RANDOM_FLOAT,RANDOM_FLOAT); 
		glBegin(GL_POLYGON);
		glNormal3f(CGAL::to_double(normalsVector.x()), CGAL::to_double(normalsVector.y()), CGAL::to_double(normalsVector.z())); 
		do {
			Point_3 pnt = j->vertex()->point();
			glVertex3d(CGAL::to_double(pnt.x()), CGAL::to_double(pnt.y()), CGAL::to_double(pnt.z())); 
		} while ( ++j != pFacet->facet_begin());
		glEnd();
	}
}

void drawCGALLine(Vertex_handle v1, Vertex_handle v2) { 
	Point_3 p1 = v1->point(); 
	Point_3 p2 = v2->point(); 
	glBegin(GL_LINES);
	glVertex3d(CGAL::to_double(p1[0]),CGAL::to_double(p1[1]),CGAL::to_double(p1[2]));
	glVertex3d(CGAL::to_double(p2[0]),CGAL::to_double(p2[1]),CGAL::to_double(p2[2]));
	glEnd(); 
}

//Strict debug function to only draw the bridge/pocket which has the longest pocket
void drawLongestBridge() { 
#ifdef DRAW_LONGEST_BRIDGE
	Bridge * curBridge = longestBridge; 

	glColor3f(1.0, 0, 1.0); //draw lines as blue
	drawCGALLine(curBridge->oPolyVertexH1, curBridge->oPolyVertexH2);

	//second draw the pocket (iterate over all vertices of pocket, and draw lines connecting them)
	//BEGIN DRAWING POCKET

	glColor3f(0, 1.0, 1.0); //draw pocket lines green
	std::list<VERTEX_TYPE>::iterator pocketIt(curBridge->pocketVertices.begin());
	Vertex_handle prevVertex, curVertex;		

	/**
	*Version below includes end points (the two nodes of the bridge)
	*/
	prevVertex = curBridge->oPolyVertexH2; 
	while (pocketIt != curBridge->pocketVertices.end()) {
		curVertex = *pocketIt; 
		//draw line from prevVertex to curVertex;
		drawCGALLine(prevVertex, curVertex); 
		prevVertex = curVertex; 
		pocketIt++; 
	}
	drawCGALLine(prevVertex, curBridge->oPolyVertexH1); 
#endif
}

void drawBridges() { 
	std::list<Bridge *>::iterator cur(bridgeEdgesList.begin());
	glLineWidth(3.0); 
	
	while (cur != bridgeEdgesList.end()) {
		Bridge * curBridge = (*cur); 

		//first draw the bridge
#ifdef DRAW_BRIDGES
		glColor3f(0, 0, 1.0); //draw lines as blue
		drawCGALLine(curBridge->oPolyVertexH1, curBridge->oPolyVertexH2);
#endif

		//second draw the pocket (iterate over all vertices of pocket, and draw lines connecting them)
		//BEGIN DRAWING POCKET
#ifdef DRAW_POCKETS
		glColor3f(0, 1.0, 0); //draw pocket lines green
		std::list<VERTEX_TYPE>::iterator pocketIt(curBridge->pocketVertices.begin());
		Vertex_handle prevVertex, curVertex;
		

		/**
		 *Version below includes end points (the two nodes of the bridge)
		 */
		prevVertex = curBridge->oPolyVertexH2; 
		while (pocketIt != curBridge->pocketVertices.end()) {
			curVertex = *pocketIt; 
			//draw line from prevVertex to curVertex;
			drawCGALLine(prevVertex, curVertex); 
			prevVertex = curVertex; 
			pocketIt++; 
		}
		drawCGALLine(prevVertex, curBridge->oPolyVertexH1); 
		
		//FINISHED DRAWING BRIDGES
#endif		
		cur++; 
	}

	drawLongestBridge(); 
}

void drawAxes(void)
{
	/* Name-stack manipulation for the purpose of
	selection hit processing when mouse button
	is pressed.  Names are ignored in normal
	OpenGL rendering mode.                    */
	Point_3 testPoint;
	//glPushMatrix();

	//BEGIN DRAWING POINTS
	/*
	glBegin(GL_POINTS); 
	glColor3f(1,1,1);
	glPointSize(3.0); 
	for (int i=0; i<points.size(); i++) { 
		double x,y,z,w;

		x =  CGAL::to_double(points[i].x());
		y =  CGAL::to_double(points[i].y());
		z =  CGAL::to_double(points[i].z()); 

		glVertex3d(x,y,z); 
		//std::cout << i << ": " << points[i] << std::endl; 
	} 
	glEnd();
	*/ 
	//END DRAWING POINTS
	drawPolyhedron(&polyhedron);

	drawBridges(); 
	/*
	glColor3f(0.3,0.3,0.3);
	glutSolidSphere(0.7, 20, 20);

	glPushMatrix();
	glPushName(1);            
	glColor3f(1,0,0);
	glRotatef(90,0,1,0);
	glutSolidCone(0.6, 4.0, 20, 20);
	glPopName();
	glPopMatrix();

	glPushMatrix ();
	glPushName(2);            
	glColor3f(0,1,0);
	glRotatef(-90,1,0,0);
	glutSolidCone(0.6, 4.0, 20, 20);
	glPopName();
	glPopMatrix();

	glColor3f(0,0,1);         
	glPushName(3);
	glutSolidCone(0.6, 4.0, 20, 20);
	glPopName();
	*/
	//glPopMatrix();
}

/* Callback function for drawing */

void display(void)
{
   //GLERROR;

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   drawAxes();
   glutSwapBuffers();

   //GLERROR;
}

/* Callback function for pick-event handling from ZPR */

void pick(GLint name)
{
	/*
   printf("Pick: %d\n",name);
   fflush(stdout);
   */ 
}

static GLfloat light_ambient[]  = { 1.0, 1.0, 1.0, 1.0 };//{ 0.0, 0.0, 0.0, 1.0 };
static GLfloat light_diffuse[]  = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat light_specular[] = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat light_position[] = { 5.0, 5.0, 100.0, 0.0 };

static GLfloat mat_ambient[]    = { 0.7, 0.7, 0.7, 1.0 };
static GLfloat mat_diffuse[]    = { 0.8, 0.8, 0.8, 1.0 };
static GLfloat mat_specular[]   = { 1.0, 1.0, 1.0, 1.0 };
static GLfloat high_shininess[] = { 100.0 };

/* Entry point */

void initOpenGL(int argc, char **argv)
{
    /* Initialise GLUT and create a window */

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize(640,640);
    glutCreateWindow("ACD Viewer");

    /* Configure GLUT callback functions */

    glutDisplayFunc(display);

    //glScalef(0.25,0.25,0.25);
	float scale = 0.10; 
	glScalef(scale,scale,scale);

    /* Configure ZPR module */

    zprInit();
    zprSelectionFunc(drawAxes);     /* Selection mode draw function */
    zprPickFunc(pick);              /* Pick event client callback   */

     /* Initialise OpenGL */

    GLERROR;

    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);

    glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT, GL_SHININESS, high_shininess);

	//added by me
	glEnable(GL_CULL_FACE);
	glColorMaterial(GL_FRONT,GL_AMBIENT_AND_DIFFUSE);
	glFrontFace(GL_CCW);
	//end of me

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_NORMALIZE);
    glEnable(GL_COLOR_MATERIAL);

    GLERROR;

    /* Enter GLUT event loop */

    glutMainLoop();

    return;
}
