#ifndef CGAL_JAVAVIEW_STREAM_H
#define CGAL_JAVAVIEW_STREAM_H

#include <utility>
#include <string>

CGAL_BEGIN_NAMESPACE

/** An object of the class Javaview_stream is a stream in which
    geometric objects can be inserted via the class
    Javaview_writer. The constructor forks a process and establishes
    two pipes between the processes. The forked process is then
    overlaid with Javaview. The file descriptors stdin and stdout of
    Javaview are hooked on the two pipes.
 */
class Javaview_stream {
public:
    Javaview_stream();
    ~Javaview_stream();

    Javaview_stream &operator<<(const std::string & s);
    Javaview_stream &operator<<(int i);
    Javaview_stream &operator<<(unsigned int i);
    Javaview_stream &operator<<(long i);
    Javaview_stream &operator<<(unsigned long i);
    Javaview_stream &operator<<(double d);

    /** Enable to trace the produced output to Javaview on the
	std::cerr (just designed for debugging). 
    */
    bool set_trace(bool b)
    {
	std::swap(b, trace_flag);
	return b;
    }
    bool get_trace() const
    {
	return trace_flag;
    }

    void trace(const std::string s) const
    {
        if (get_trace())
            std::cerr << s;
    }
    void trace(double d) const
    {
        if (get_trace())
            std::cerr << d << ' ';
    }
    void trace(int i) const
    {
        if (get_trace())
            std::cerr << i << ' ';
    }
    void trace(unsigned int i) const
    {
        if (get_trace())
            std::cerr << i << ' ';
    }

 private:
    void setup_javaview();

    bool trace_flag;  // makes operator<<() write a trace on cerr.
    int in, out;      // file descriptors for input and output pipes
    int pid;          // the javaview process identification
};

CGAL_END_NAMESPACE

#endif // CGAL_JAVAVIEW_STREAM_H
